/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef READYAMLFILES_H_
#define READYAMLFILES_H_
#include <algorithm>
#include "ginac/ginac.h"
#include "tools.h"
#include "integral.h"
#include "yaml-cpp/yaml.h"

typedef std::vector< symmetries > SYM;
typedef SYM::iterator ItSYM;

typedef std::vector< Equation * > VE;
typedef VE::iterator ItVE;

struct Jobs {
  std::vector<std::pair<std::string,std::string> > sector2Reduce;
  std::vector<std::pair<std::string,std::string> > den;
  std::vector<std::pair<std::string,std::string> > num;
  std::vector<std::pair<std::string,std::string> > mandatoryFile;
  std::vector<std::pair<std::string,std::string> > optionalFile;
  std::vector<std::vector<std::string> > mandatoryRec;
  std::vector<std::vector<std::string> > optionalRec;
  std::string specialname;
  std::string runBacksubstitution;
  std::string runTriangular;
  std::string runPyred;
  std::string runInitiate;
  unsigned max_parallel_sectors;
  std::string runSymmetries;
  std::string dataBase;
  std::string dataFile;
  std::string runOTF;
  std::string outputDir;
  std::string serialWrite;
  std::string skipPyred;
  std::string writeNumericalSystem;
  std::multimap<std::string, std::vector<int> > selectMastersReduction;
};

struct Kira2File {
  std::vector<std::pair<std::string,std::string> > target;
  std::string inputDir;
  std::string reconstructMass;
};

struct Integral_F {
  std::string name;
  std::vector<std::string> loop;
  std::vector<GiNaC::possymbol> loopVar;
  std::vector<GiNaC::possymbol> allVar;
  GiNaC::lst loop2loop, loop2loop2, loopVarList, loopVarList2;
  GiNaC::lst props, propsMomFlowA;
  std::vector<GiNaC::lst> scal2Props, propsMomFlowB;
  int jule;
  GiNaC::possymbol *propSymb;
  GiNaC::lst scalprod;
  GiNaC::ex G, U, F; // U + F Polynom
  GiNaC::lst FPolynom;
  std::vector<int> *mask, *allowSector, *skipSector;
  std::vector<int> zeroSector;
  VE identitiesIBP, identitiesLI;
  SYM *symVec, *relVec;
  std::vector<int> sector2Reduce, bound;
  int biggestSector2Reduce;
  std::vector <unsigned> lowestSectors;
  int topology;
  std::vector <std::pair<std::string,std::string> > propagator;
  int* invarID;
  std::vector<std::string> invarMap;
  std::vector<int> topLevelSectors;
};

struct Kinematics {
  std::vector<std::string> im;
  std::vector<std::string> om;
  std::string rpl;
  std::pair<std::string,std::string> mc;
  std::vector<std::pair<std::string,int> > ki;
  std::vector <std::pair<std::pair<std::string,std::string>,std::string> > sr;
};

namespace YAML {
  template<>
  struct convert<Jobs> {
    static bool decode(const Node& node, Jobs& rhs) {
      
      
      for(YAML::const_iterator it=node.begin(); it != node.end();++it){
        std::string token = (*it).first.as<std::string>();
        if(token != "sector_selection" && token != "identities" && token != "select_integrals" 
          && token != "run_triangular" && token != "run_back_substitution" && token != "run_pyred"
          && token != "run_initiate" && token != "run_symmetries" && token != "run_on_the_fly" 
          && token != "alt_dir" && token != "data_file" && token != "data_base" 
          && token != "serial_write" && token != "skip_pyred" && token != "max_parallel_sectors" 
          && token != "write_numerical_system" && token != "select_masters_reduction") {
          std::cout << "\n***********************************************************\n";
          std::cout << "***Kira does not know the option: " << token << std::endl;
          std::cout << "***In the job file.";
          std::cout << "\n***********************************************************\n";
          exit(-1);
        }
      }
      
      if(node["sector_selection"]){ 
	
        for(YAML::const_iterator it=node["sector_selection"].begin(); it != node["sector_selection"].end();++it){
          std::string token = (*it).first.as<std::string>();
          if(token != "select_recursively") {
            std::cout << "\n***********************************************************\n";
            std::cout << "***Kira does not know the option: " << token << std::endl;
            std::cout << "***In the job file.";
            std::cout << "\n***********************************************************\n";
            exit(-1);
          }
        }
        
	if(node["sector_selection"]["select_recursively"]){
	  
	  const Node& nodeb = node["sector_selection"]["select_recursively"];
	  
	  for(unsigned i=0;i<nodeb.size();i++){
	  
	    rhs.sector2Reduce.push_back(make_pair(nodeb[i][0].as<std::string>(),nodeb[i][1].as<std::string>()));
	  }
	}
      }
      
      if(node["identities"]){
        
        for(YAML::const_iterator it=node["identities"].begin(); it != node["identities"].end();++it){
          std::string token = (*it).first.as<std::string>();
          if(token != "ibp") {
            std::cout << "\n***********************************************************\n";
            std::cout << "***Kira does not know the option: " << token << std::endl;
            std::cout << "***In the job file.";
            std::cout << "\n***********************************************************\n";
            exit(-1);
          }
        }
	
	if(node["identities"]["ibp"]){
	  
	  const Node& nodeibp = node["identities"]["ibp"];
	  
	  for(unsigned i=0;i<nodeibp.size();i++){
	  
	    rhs.den.push_back(make_pair(nodeibp[i]["r"][0].as<std::string>(),nodeibp[i]["r"][1].as<std::string>()));
	    rhs.num.push_back(make_pair(nodeibp[i]["s"][0].as<std::string>(),nodeibp[i]["s"][1].as<std::string>()));
	  }
	}
      }
      
      if(node["select_integrals"]){
        
        for(YAML::const_iterator it=node["select_integrals"].begin(); it != node["select_integrals"].end();++it){
          std::string token = (*it).first.as<std::string>();
          if(token != "select_mandatory_list" && token != "select_optional_list" && token != "select_mandatory_recursively" 
                      && token != "select_optional_recursively") {
            std::cout << "\n***********************************************************\n";
            std::cout << "***Kira does not know the option: " << token << std::endl;
            std::cout << "***In the job file.";
            std::cout << "\n***********************************************************\n";
            exit(-1);
          }
        }
      
	if(node["select_integrals"]["select_mandatory_list"]){
	  
	  const Node& nodeibp = node["select_integrals"]["select_mandatory_list"];
	  
	  for(unsigned i=0;i<nodeibp.size();i++){
	  
	    rhs.mandatoryFile.push_back(make_pair(nodeibp[i][0].as<std::string>(),nodeibp[i][1].as<std::string>()));
	  }
	}
	
	if(node["select_integrals"]["select_optional_list"]){
	  
	  const Node& nodeibp = node["select_integrals"]["select_optional_list"];
	  
	  for(unsigned i=0;i<nodeibp.size();i++){
	  
	    rhs.optionalFile.push_back(make_pair(nodeibp[i][0].as<std::string>(),nodeibp[i][1].as<std::string>()));
	  }
	}
	
	if(node["select_integrals"]["select_mandatory_recursively"]){
	  const Node& nodeibp = node["select_integrals"]["select_mandatory_recursively"];
	  for(unsigned i=0;i<nodeibp.size();i++){
	    std::vector<std::string> info;
	    info.push_back(nodeibp[i][0].as<std::string>());
	    info.push_back(nodeibp[i][1].as<std::string>());
	    info.push_back(nodeibp[i][2].as<std::string>());
	    info.push_back(nodeibp[i][3].as<std::string>());
	    rhs.mandatoryRec.push_back(info/*make_pair(nodeibp[i][0].as<std::string>()
	    ,nodeibp[i][1].as<std::string>())*/);
	  }
	}
	
	if(node["select_integrals"]["select_optional_recursively"]){
	  const Node& nodeibp = node["select_integrals"]["select_optional_recursively"];
	  for(unsigned i=0;i<nodeibp.size();i++){
	    std::vector<std::string> info;
	    info.push_back(nodeibp[i][0].as<std::string>());
	    info.push_back(nodeibp[i][1].as<std::string>());
	    info.push_back(nodeibp[i][2].as<std::string>());
	    info.push_back(nodeibp[i][3].as<std::string>());
	    
	    rhs.optionalRec.push_back(info/*make_pair(nodeibp[i][0].as<std::string>(),nodeibp[i][1].as<std::string>())*/);
	  }
	}
      }
      
      if(node["select_masters_reduction"]){
        
        const Node& nodeTop = node["select_masters_reduction"];
        
        for(unsigned iM=0;iM<nodeTop.size();iM++) {
        
          if(nodeTop[iM].size()!=2){
            std::cout << "Wrong options in:\n";
            std::cout << "select_masters_reduction\n";
            continue;
          }
          std::vector<int> token;
          for(unsigned jM=0; jM<nodeTop[iM][1].size();jM++) {
            token.push_back(nodeTop[iM][1][jM].as<int>());
          }
          rhs.selectMastersReduction.insert(std::make_pair(nodeTop[iM][0].as<std::string>(),token));
        }
      }
      
      if (node["run_triangular"])
        rhs.runTriangular=node["run_triangular"].as<std::string>();
      if (node["run_back_substitution"])
        rhs.runBacksubstitution=node["run_back_substitution"].as<std::string>();
      if (node["run_pyred"])
        rhs.runPyred=node["run_pyred"].as<std::string>();
      if (node["run_initiate"])
        rhs.runInitiate=node["run_initiate"].as<std::string>();
      rhs.max_parallel_sectors=0;
      if (node["max_parallel_sectors"])
        rhs.max_parallel_sectors=node["max_parallel_sectors"].as<unsigned>();
      if (node["run_symmetries"])
        rhs.runSymmetries=node["run_symmetries"].as<std::string>();
      if (node["run_on_the_fly"])
        rhs.runOTF=node["run_on_the_fly"].as<std::string>();
      if (node["data_base"])
        rhs.dataBase=node["data_base"].as<std::string>();
      if (node["data_file"])
        rhs.dataFile=node["data_file"].as<std::string>();
      if (node["alt_dir"])
        rhs.outputDir=node["alt_dir"].as<std::string>();
      if (node["serial_write"])
        rhs.serialWrite=node["serial_write"].as<std::string>();
      if (node["skip_pyred"])
        rhs.skipPyred=node["skip_pyred"].as<std::string>();
      if (node["write_numerical_system"])
        rhs.writeNumericalSystem=node["write_numerical_system"].as<std::string>();
      
      return true;
    }
  };
  
  
  template<>
  struct convert<Kira2File> {
    static bool decode(const Node& node, Kira2File& rhs) {
      
      for(YAML::const_iterator it = node.begin(); it != node.end(); it++){
        std::string token = (*it).first.as<std::string>();
        if(token != "target" && token != "reconstruct_mass" && token != "alt_dir") {
          std::cout << "\n***********************************************************\n";
          std::cout << "***Kira does not know the option: " << token << std::endl;
          std::cout << "***In the job file.";
          std::cout << "\n***********************************************************\n";
        }
      }
      
      if(node["target"]){
	  
        const Node& nodet = node["target"];
        
        for(unsigned i=0;i<nodet.size();i++){
          if(nodet[i].size()!=2){
            std::cout << "Wrong options in:\n";
            std::cout << "target: [topo,seeds]\n";
            continue;
          }
          rhs.target.push_back(make_pair(nodet[i][0].as<std::string>(),nodet[i][1].as<std::string>()));
        }
      }
   
      if (node["alt_dir"])
        rhs.inputDir=node["alt_dir"].as<std::string>();
      if (node["reconstruct_mass"])
        rhs.reconstructMass=node["reconstruct_mass"].as<std::string>();
      
      return true;
    }
  };


  template<>
  struct convert<Kinematics> {
    static bool decode(const Node& node, Kinematics& rhs) {
      
      for(YAML::const_iterator it=node.begin(); it != node.end();++it){
        std::string token = (*it).first.as<std::string>();
        if(token != "incoming_momenta" && token != "outgoing_momenta" && token != "momentum_conservation" 
          && token != "kinematic_invariants" && token != "scalarproduct_rules" && token != "symbol_to_replace_by_one") {
          std::cout << "Kira does not know the option: " << token << std::endl;
          std::cout << "In config/kinematics.yaml" << std::endl;
          exit(-1);
        }
      }
      
      if(node["incoming_momenta"]){
        const Node& nodei = node["incoming_momenta"];
        for(unsigned i=0;i<nodei.size();i++) {
          rhs.im.push_back(nodei[i].as<std::string>());
        }
      }
      if(node["outgoing_momenta"]){
        const Node& nodeo = node["outgoing_momenta"];
        for(unsigned i=0;i<nodeo.size();i++) {
          rhs.om.push_back(nodeo[i].as<std::string>());
        }
      }
      
      if(node["momentum_conservation"]){
        const Node& nodem = node["momentum_conservation"];
        if (nodem.size())
          rhs.mc=make_pair(nodem[0].as<std::string>(),nodem[1].as<std::string>());
      }
      
      if(node["kinematic_invariants"]){
        const Node& nodek = node["kinematic_invariants"];
        for(unsigned i=0;i<nodek.size();i++) {
          rhs.ki.push_back(std::pair<std::string, int > (nodek[i][0].as<std::string>(),
                                              nodek[i][1].as<int>()));
        }
      }
      if(node["scalarproduct_rules"]){
        const Node& nodes = node["scalarproduct_rules"];
        for(unsigned i=0;i<nodes.size();i++) {
          std::pair<std::string,std::string> pure(nodes[i][0][0].as<std::string>(),
                                  nodes[i][0][1].as<std::string>());
          rhs.sr.push_back(std::pair<std::pair<std::string,std::string>,std::string>
                          (pure,nodes[i][1].as<std::string>()));
        }
      }
      
      if(node["symbol_to_replace_by_one"]){
	const Node noderpl=node["symbol_to_replace_by_one"];
	rhs.rpl=noderpl.as<std::string>();
      }
      
      return true;
    }
  };

  template<>
  struct convert<Integral_F> {
    static bool decode(const Node& node, Integral_F& rhs) {

      
      for(YAML::const_iterator it=node.begin(); it != node.end();++it){
        std::string token = (*it).first.as<std::string>();
        if(token != "propagators" && token != "loop_momenta" && token != "name" && token != "top_level_sectors") {
          std::cout << "Kira does not know the option: " << token << std::endl;
          std::cout << "In config/integralfamilies.yaml" << std::endl;
          exit(-1);
        }
      }
      
      if(!node["propagators"]){
        std::cout << "Option propagators is missing in"
        << "config/integralfamilies.yaml" << std::endl;
        exit(-1);
      }
      if(!node["loop_momenta"]){
        std::cout << "Option loop_momenta is missing in"
        << "config/integralfamilies.yaml" << std::endl;
        exit(-1);
      }
      if(!node["name"]){
        std::cout << "Option name is missing in"
        << "config/integralfamilies.yaml" << std::endl;
        exit(-1);
      }
      
      rhs.name = node["name"].as<std::string>();
      
      const Node& nodet = node["loop_momenta"];
      for(unsigned i=0;i<nodet.size();i++) {
	rhs.loop.push_back(nodet[i].as<std::string>());
	rhs.allVar.push_back(get_symbol(nodet[i].as<std::string>()));
	rhs.loopVar.push_back(get_symbol(nodet[i].as<std::string>()));
	rhs.loopVarList.append(get_symbol(nodet[i].as<std::string>()));
	
	std::string loop2 = rhs.loop[i] + "2";
        rhs.loopVarList2.append(get_symbol(loop2));
	
        rhs.loop2loop2.append(rhs.loopVar[i]==get_symbol(loop2)); // k == k'
	rhs.loop2loop.append(get_symbol(loop2)==rhs.loopVar[i]); // k' == k
      }
      
      const Node& nodep = node["propagators"];
      for(unsigned i=0;i<nodep.size();i++) {
        rhs.propagator.push_back( std::pair<std::string, std::string >
                                ( nodep[i][0].as<std::string>(),
                                  nodep[i][1].as<std::string>() ) );
      }
      
      rhs.jule=static_cast<int>(nodep.size());
      
      int usedOption = 0;
      
      if(node["top_level_sectors"]){
        
        const Node& nodeTop = node["top_level_sectors"];
        
        for(unsigned i=0;i<nodeTop.size();i++) {
        
          rhs.topLevelSectors.push_back(nodeTop[i].as<int>());
        }
        usedOption = 1;
      }
      
      if(usedOption == 0)
	rhs.topLevelSectors.push_back((1<<rhs.jule)-1);
	
      return true;
    }
  };

  template<>
  struct convert<IBPIntegral> {
    static bool decode(const Node& node, IBPIntegral& rhs) {

      const Node& noderef = node;
      
      rhs.id = noderef[0].as<int>();
      rhs.flag2 = noderef[1].as<int>();
      rhs.characteristics[SECTOR] = noderef[2].as<int>();
      rhs.characteristics[DOTS] = noderef[3].as<int>();
      rhs.characteristics[NUM] = noderef[4].as<int>();
      rhs.length = noderef[5].as<int>();
      for(unsigned i=6;i<noderef.size()-1;i++){
        rhs.indices[i-6]=noderef[i].as<int>();
      }
      rhs.coefficientString = noderef[noderef.size()-1].as<std::string>();
      return true;
      
    }
  };
}

#endif 
