/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#include <sstream>
#include <algorithm>

#include "coeff_vec.h"
#include "coeff_helper.h" // mod_mul, mod_pow

namespace pyred
{

// static members
bool Coeff_vec::auto_vars{true};
std::string Coeff_vec::set_to_one{};
uint Coeff_vec::ncoeffs{2};
cvec Coeff_vec::prime{primes().at(0), primes().at(1)};
std::vector<std::pair<std::string,cvec>> Coeff_vec::randvars{};

void Coeff_vec::init(const std::vector<std::string>& vars, uint n,
                     const std::string& set1) {
  if (n < 1) {
    std::stringstream ss;
    ss << "Invalid number of coefficients in vector: " << n << std::endl;
    throw input_error(ss.str());
  }
  if (!vars.empty()) {
    // If at lease one variable was set manually,
    // turn off auto variable declaration.
    auto_vars = false;
    if (!set1.empty() &&
        std::find(vars.cbegin(), vars.cend(), set1) == vars.cend()) {
      // If set1 is set and auto varible declaration is off,
      // set1 must be a declared variable.
      std::stringstream ss;
      ss << "'" << set1 << "' is not declared as a variable." << std::endl;
      throw input_error(ss.str());
    }
  }
  set_to_one = set1;
  ncoeffs = n;
  prime.clear();
  prime.reserve(n);
  auto pit = primes().cbegin();
  for (uint k = 0; k < n; ++k) {
    prime.emplace_back(*pit);
    // If more coefficients are requested as primes are available,
    // use primes multiply with different random numbers.
    if (++pit == primes().cend()) pit = primes().cbegin();
  }
  randvars.clear();
  randvars.reserve(vars.size());
  for (const auto& var: vars) {
    randvars.push_back({var, std::vector<uint64_t>()});
    randvars.back().second.reserve(n);
  }
  for (const auto& p: prime) {
    for (auto& vv: randvars) {
      if (vv.first == set1) {
        vv.second.push_back(1);
      }
      else {
        vv.second.push_back(random_minmax(randvar_lbound(), p));
      }
    }
  }
  return;
}

Coeff_vec::Coeff_vec(): c(cvec()) {}
Coeff_vec::Coeff_vec(const Coeff_vec& a): c(a.c) {}

Coeff_vec::Coeff_vec(const std::string& s) {
  for (const auto& vc: randvars) {
    if (vc.first == s) {
      c = vc.second;
      return;
    }
  }
  std::stringstream ss(s);
  uint64_t cc;
  auto success = bool(ss >> cc);
  if (!(success && ss.rdbuf()->in_avail() == 0)) {
    if (!auto_vars) {
      throw parser_error(s);
    }
    // auto_vars: add new symbols automatically and assign values to them.
    if (!is_valid_varname(s)) {
      throw parser_error(s + std::string(
        "; symbols are expected to be of the form [A-Za-z_][A-Za-z0-9_]*."));
    }
    for (const auto& p: prime) {
      if (s == set_to_one) {
        c.push_back(1);
      }
      else {
        c.push_back(random_minmax(randvar_lbound(), p));
      }
    }
    randvars.push_back({s, c});
    // std::cout << "define symbol " << s << std::endl;
  }
  else {
    c = cvec(ncoeffs, cc);
  }
}

Coeff_vec& Coeff_vec::operator+=(const Coeff_vec& a) {
  auto ait = a.c.cbegin();
  auto pit = prime.cbegin();
  for (auto& cc: c) {
    cc += *ait++;
    if (cc >= *pit) cc -= *pit;
    ++pit;
  }
  return *this;
}
Coeff_vec& Coeff_vec::operator-=(const Coeff_vec& a) {
  auto ait = a.c.cbegin();
  auto cit = c.begin();
  for (const auto& p: prime) {
    if (*ait > *cit) *cit += p;
    *cit++ -= *ait++;
  }
  return *this;
}
Coeff_vec& Coeff_vec::operator*=(const Coeff_vec& a) {
  auto ait = a.c.cbegin();
  auto pit = prime.cbegin();
  for (auto& cc: c) {
    cc = mod_mul(cc, *ait++, *pit++);
  }
  return *this;
}
Coeff_vec& Coeff_vec::operator/=(const Coeff_vec& a) {
  auto ait = a.c.cbegin();
  auto pit = prime.cbegin();
  for (auto& cc: c) {
    if (*ait == uint64_t(0)) {
      throw zero_modular_division();
    }
    cc = mod_mul(cc, mod_inv(*ait++, *pit), *pit);
    ++pit;
  }
  return *this;
}
Coeff_vec& Coeff_vec::operator^=(const Coeff_vec& a) {
  // assume that exponents are small integers
  uint64_t exp;
  uint64_t base;
  auto ait = a.c.cbegin();
  auto pit = prime.cbegin();
  for (auto& cc: c) {
    if (*ait == 0) {
      exp = 0;
      cc = 1;
    }
    else {
      if (2*(*ait) < *pit) {
        // treat as positive exponent
        exp = *ait;
        base = cc;
      }
      else {
        // treat as negative exponent
        exp = *pit - *ait;
        base = mod_pow(cc, (*pit)-uint64_t(2), *pit); // =1/cc
        cc = base;
      }
      for (uint64_t i = 1; i < exp; ++i) {
        cc = mod_mul(cc, base, *pit);
      }
    }
    ++ait;
    ++pit;
  }
  return *this;
}

Coeff_vec operator-(const Coeff_vec& a) {
  Coeff_vec cv;
  cv.c.reserve(Coeff_vec::ncoeffs);
  auto pit = Coeff_vec::prime.cbegin();
  for (const auto& aa: a.c) {
    if (aa == 0) {
      cv.c.emplace_back(aa);
      ++pit;
    }
    else {
      cv.c.emplace_back(*pit++ - aa);
    }
  }
  return cv;
}
Coeff_vec operator+(const Coeff_vec& a, const Coeff_vec& b) {
  Coeff_vec cv;
  cv.c.reserve(Coeff_vec::ncoeffs);
  auto ait = a.c.cbegin();
  auto bit = b.c.cbegin();
  for (const auto& p: Coeff_vec::prime) {
    auto sum = *ait++ + *bit++;
    if (sum >= p) sum -= p;
    cv.c.emplace_back(sum);
  }
  return cv;
}
Coeff_vec operator-(const Coeff_vec& a, const Coeff_vec& b) {
  Coeff_vec cv;
  cv.c.reserve(Coeff_vec::ncoeffs);
  auto ait = a.c.cbegin();
  auto pit = Coeff_vec::prime.cbegin();
  for (const auto& bb: b.c) {
    auto diff = *ait++;
    if (bb > diff) diff += *pit++;
    cv.c.emplace_back(diff-bb);
  }
  return cv;
}
Coeff_vec operator*(const Coeff_vec& a, const Coeff_vec& b) {
  Coeff_vec cv;
  cv.c.reserve(Coeff_vec::ncoeffs);
  auto ait = a.c.cbegin();
  auto bit = b.c.cbegin();
  for (const auto& p: Coeff_vec::prime) {
    cv.c.emplace_back(mod_mul(*ait++, *bit++, p));
  }
  return cv;
}
Coeff_vec operator/(const Coeff_vec& a, const Coeff_vec& b) {
  Coeff_vec cv;
  cv.c.reserve(Coeff_vec::ncoeffs);
  auto ait = a.c.cbegin();
  auto bit = b.c.cbegin();
  for (const auto& p: Coeff_vec::prime) {
    if (*bit == uint64_t(0)) {
      throw zero_modular_division();
    }
    cv.c.emplace_back(mod_mul(*ait++, mod_inv(*bit++, p), p));
  }
  return cv;
}

bool operator==(const Coeff_vec& a, const Coeff_vec& b) {
  return (a.c == b.c);
}

std::ostream& operator<<(std::ostream& out, const Coeff_vec& a) {
  out << "{";
  if (!a.c.empty() > 0) {
    out << a.c.front();
  }
  for (auto it = a.c.cbegin()+1; it < a.c.cend(); ++it) {
    out << "," << *it;
  }
  out << "}";
  return out;
}

}
