/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/

#include "tools.h"
#include "dataBase.h"

using namespace std;
using namespace GiNaC;

static Loginfo & logger = Loginfo::instance();

bool sortDBintegral(const DBintegral& l, const  DBintegral& r) {
  
  if( l.id > r.id )
    return true;
  else if( l.id < r.id )
    return false;
  
  return false;
  
};

DataBase::DataBase(string name){
  errorMessage=0;
  rc = sqlite3_open(name.c_str(), &db);
  if( rc ){
    fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(db));
    sqlite3_close(db);
    exit(1);
  }else{
    fprintf(stderr, "\nOpened database successfully\n");
  }
};

DataBase::~DataBase(){
  
  sqlite3_close_v2(db);
  
};


int DataBase::checkTable() {
  
  string sql = "select 1 from sqlite_master where type='table' and tbl_name='EQUATION'";

  sqlite3_prepare_v2(db, sql.c_str(), sql.size(), &stmt, NULL);

  sqlite3_step(stmt);

  int token = something_int(sqlite3_column_text(stmt, 0));

  sqlite3_finalize(stmt);
  return token;
}


void DataBase::create_integral_table(int lengthIndices){

  string sql = "CREATE TABLE INTEGRAL(" \
          "ID            TEXT   NOT NULL,";
  sql +=  "INDICES       TEXT   NOT NULL," \
          "SECTOR        INT    NOT NULL," \
          "TOPOLOGY      INT    NOT NULL," \
          "DENCOUNT      INT    NOT NULL," \
          "DOTS          INT    NOT NULL," \
          "NUMS          INT    NOT NULL," \
          "AUXILIARY     INT    NOT NULL," \
          "UNIQUE (ID)," \
          "PRIMARY KEY (INDICES,TOPOLOGY));";
 
  columnName.push_back("ID");
  columnName.push_back("INDICES");
  columnName.push_back("SECTOR");
  columnName.push_back("TOPOLOGY");
  columnName.push_back("DENCOUNT");
  columnName.push_back("DOTS");
  columnName.push_back("NUMS");
  columnName.push_back("AUXILIARY");

//   totalColumn=NDBINT;
//   if(totalColumn!=columnName.size()){
//     logger << __FILE__ << "something is wrong\n";
//     exit(-1);
//   }
  
  execute_statement(sql,"Integral table created successfully\n");
  
};

void DataBase::create_equation_table(){
  
  string sql = "CREATE TABLE EQUATION("  \
      "ID              TEXT    NOT NULL," \
      "ARRAYID         TEXT    NOT NULL," \
      "AUXILIARY       INT    NOT NULL," \
      "SECTOR          INT    NOT NULL," \
      "LENGTH          INT    NOT NULL," \
      "COEFFICIENT     TEXT   NOT NULL," \
      "PRIMARY KEY (ID,ARRAYID));";
  
  execute_statement(sql,"Equation table created successfully\n");
  
}

void DataBase::execute_statement(string& sql,string message){
  
  rc = sqlite3_exec(db, sql.c_str(), NULL, NULL, &errorMessage);
  
  if( rc != SQLITE_OK ){
    fprintf(stderr, "SQLite3: %s\n", errorMessage);
    sqlite3_free(errorMessage);
  }
  else{
    logger << message;
  }
  
};

void DataBase::begin_transaction(){
  
  string sql = "BEGIN TRANSACTION";
  rc = sqlite3_exec(db, sql.c_str(), NULL, NULL, &errorMessage);
  
  if( rc != SQLITE_OK ){
    fprintf(stderr, "SQL error: %s\n", errorMessage);
    sqlite3_free(errorMessage);
  }
//   else{
//     logger<< message;
//   }
};

void DataBase::commit_transaction(){
  
  string sql = "COMMIT TRANSACTION";
  rc = sqlite3_exec(db, sql.c_str(), NULL, NULL, &errorMessage);
  
  if( rc != SQLITE_OK ){
    fprintf(stderr, "SQL error: %s\n", errorMessage);
    sqlite3_free(errorMessage);
  }
//   else{
//     logger<< message;
//   }

  sqlite3_finalize(stmt);
  
}

void DataBase::prepare_integral(){

  string input = "INSERT INTO INTEGRAL VALUES (";
  
  int jj=1;
  for(int i=0; i<(NDBINT-1);i++){
  
    input+="?"+int_string(jj++)+",";
  }
  
  input+="?"+int_string(jj++)+")";
  
  sqlite3_prepare_v2(db, input.c_str(), input.size(), &stmt, NULL);

};


void DataBase::prepare_pyred(){
  
  string input = "SELECT ID FROM INTEGRAL WHERE ";
  
//   input+=columnName[0];//id
//   input+="=?"+int_string(1);
  
//   input+=" AND ";
  
  input+=columnName[3];//topology
  input+="=?"+int_string(1);
  
  input+=" AND ";
  
//   input+=columnName[2];//sector
//   input+="&?"+int_string(2) + "="+columnName[2];

  input+=columnName[2];//sector
  input+=" & ~?"+int_string(2)+"==0";
  
  
  input+=" AND ";
  
  input+=columnName[5];//r
//   input+="-";
//   input+=columnName[4];//t
  input+="<=?"+int_string(3);
  
  input+=" AND ";
  
  input+=columnName[6];//s
  input+="<=?"+int_string(4);
  
  input+=" AND ";
  
  input+=columnName[7];//auxiliary
  input+="=?"+int_string(5);
  
  cout << input << endl;
  sqlite3_prepare_v2(db, input.c_str(), input.size(), &stmt, NULL);
  
}


void DataBase::prepare_id(){
  
  string input = "SELECT * FROM INTEGRAL WHERE ";
  
  
  input+=columnName[1];
  input+="=?"+int_string(1);
  input+=" AND ";
  input+=columnName[3];
  input+="=?"+int_string(2);
  
  sqlite3_prepare_v2(db, input.c_str(), input.size(), &stmt, NULL);
  
}

void DataBase::prepare_lookup_id(){

  string input = "SELECT * FROM EQUATION INNER JOIN INTEGRAL ON EQUATION.ARRAYID=INTEGRAL.ID WHERE EQUATION.ID=?";
  
  sqlite3_prepare_v2(db, input.c_str(), input.size(), &stmt, NULL);
  
}

void DataBase::prepare_backsubstitution(){

  string input = "INSERT INTO EQUATION VALUES (?1, ?2, ?3, ?4, ?5, ?6)";

  sqlite3_prepare_v2(db, input.c_str(), input.size(), &stmt, NULL);
  
}

void DataBase::bind_integral(SEEDIntegral& integral, int lengthIndices){
  
  sqlite3_bind_text(stmt, 1, to_string(integral.id).c_str(), -1, SQLITE_TRANSIENT);
  
  string indices;
  for(int i=0; i<lengthIndices; i++){
    indices+=int_string(integral.indices[i]);
    if(i!=lengthIndices-1)
      indices+=",";
  }
  sqlite3_bind_text(stmt, 2, indices.c_str(), indices.size(), SQLITE_TRANSIENT);
    
  sqlite3_bind_int(stmt, 3, integral.characteristics[SECTOR]);
  sqlite3_bind_int(stmt, 4, integral.characteristics[TOPOLOGY]);
  sqlite3_bind_int(stmt, 5, integral.characteristics[DENCOUNT]);
  sqlite3_bind_int(stmt, 6, integral.characteristics[DOTS]);
  sqlite3_bind_int(stmt, 7, integral.characteristics[NUM]);
  sqlite3_bind_int(stmt, 8, integral.flag2);
    
  sqlite3_step(stmt);
  sqlite3_reset(stmt);
};

void DataBase::bind_equation(Integral& integral, int length, long unsigned ID){
  
  sqlite3_bind_text(stmt, 1, to_string(ID).c_str(), -1, SQLITE_TRANSIENT);
  sqlite3_bind_text(stmt, 2, to_string(integral.id).c_str(), -1, SQLITE_TRANSIENT);
  sqlite3_bind_int(stmt, 3, integral.flag2);
  sqlite3_bind_int(stmt, 4, integral.characteristics[SECTOR]);
  sqlite3_bind_int(stmt, 5, length);
  sqlite3_bind_text(stmt, 6, integral.coefficientString.c_str(),
		integral.coefficientString.size(), SQLITE_STATIC);
  
  sqlite3_step(stmt);

  if (sqlite3_step(stmt) != SQLITE_DONE) {
    sqlite3_free(errorMessage);
  }

  sqlite3_reset(stmt);
}

int DataBase::number_of_columns(){
  
  string sql = "SELECT * FROM INTEGRAL;";
  
  sqlite3_prepare_v2(db, sql.c_str(), sql.size(), &stmt, NULL);
  
  totalColumn = sqlite3_column_count(stmt);
  
  sqlite3_step(stmt);
  
  for(unsigned int iterator = 0; iterator<totalColumn; iterator++){
    columnName.push_back(sqlite3_column_name(stmt, iterator));
    sqlite3_reset(stmt);
  }

  sqlite3_finalize(stmt);
  if(totalColumn==0)
    logger << "The database is yet empty, skip it\n";  
  return totalColumn;
  
};

void DataBase::prepare_find_master(){
  
  string input = "SELECT * FROM INTEGRAL WHERE ";
  
  input+=columnName[0];//id
  input+="=?"+int_string(1);
    
  sqlite3_prepare_v2(db, input.c_str(), input.size(), &stmt, NULL);
  
}

int DataBase::bind_id_get_integral(long unsigned id, tuple<string,unsigned, unsigned>& integral){
  
  sqlite3_bind_text(stmt, 1, to_string(id).c_str(), -1, SQLITE_STATIC);
  
  int res = sqlite3_step(stmt);
  if ( res == SQLITE_ROW ) {
    get<0>(integral) = (char*)sqlite3_column_text(stmt, 1);
    
    unsigned token;
    std::string unsignedLong = (char*)sqlite3_column_text(stmt, 3);
    std::istringstream ss(unsignedLong);
    ss >> token;
    get<1>(integral) = token;
    sqlite3_reset(stmt);
    return 1;
  }
  if ( res == SQLITE_DONE || res==SQLITE_ERROR) {
    sqlite3_reset(stmt);
    return 0;
  }
  return 0;
}

void DataBase::bind_id_get_equation(long unsigned id, vector<DBintegral>& equation){
  
  sqlite3_bind_text(stmt, 1, to_string(id).c_str(), -1, SQLITE_TRANSIENT);
  
  while(1){
    int res = sqlite3_step(stmt);
    if ( res == SQLITE_ROW ) {
      DBintegral integral;
      integral.indices = (char*)sqlite3_column_text(stmt, 7);
      int massDimension=0;
      size_t found;
      int indices;
      string line = integral.indices;
      while((found = line.find_first_of(",")) != string::npos){
        istringstream(line.substr(0,found)) >> indices;
        line=line.substr(found+1);
        massDimension+=indices;
      }
      istringstream (line) >> indices;
      massDimension+=indices;
      
      integral.id = std::stoul((char*)sqlite3_column_text(stmt, 1));
      
      integral.massDimension=massDimension;
      integral.topology=something_int(sqlite3_column_text(stmt,  9  ));
      integral.coefficient=(char*)sqlite3_column_text(stmt, 5);
      equation.push_back(integral);
    }
    if ( res == SQLITE_DONE || res==SQLITE_ERROR) {
	break;
    }
  }
  sort(equation.begin(),equation.end(),sortDBintegral);
  
  sqlite3_reset(stmt);
  
};

void DataBase::finalize(){
  
  sqlite3_finalize(stmt);

}


void DataBase::bind_get_answer(int topology, int sector, int dots, int nums,std::vector<long unsigned>& mandatory){
  
  sqlite3_bind_int(stmt, 1, topology);
  sqlite3_bind_int(stmt, 2, sector);
  sqlite3_bind_int(stmt, 3, dots);
  sqlite3_bind_int(stmt, 4, nums);
  sqlite3_bind_int(stmt, 5, 0);
  
  while(1){
    int res = sqlite3_step(stmt);
    
    if ( res == SQLITE_ROW ) {
      
      unsigned long token;
      std::string unsignedLong = (char*)sqlite3_column_text(stmt, 0);
      std::istringstream ss(unsignedLong);
      ss >> token;
      mandatory.push_back(token);
    }
    if ( res == SQLITE_DONE || res==SQLITE_ERROR) {
	break;
    }
  }
  
  sqlite3_reset(stmt);
  
}

long unsigned DataBase::bind_get_id(vector<int> & indices, int topologyNumber){
  
  string indicesSTRING;
  for(unsigned int iterator=0; iterator<indices.size();iterator++){
    indicesSTRING += int_string(indices[iterator]);
    if(iterator != indices.size()-1)
      indicesSTRING += ",";
  }
  sqlite3_bind_text(stmt, 1, indicesSTRING.c_str(),indicesSTRING.size(), SQLITE_STATIC);
  
  sqlite3_bind_int(stmt, 2, topologyNumber);
  
  int row=sqlite3_step(stmt); 
  
  if ( row == SQLITE_DONE || row==SQLITE_ERROR ) {
    logger << "This integral is not inside the range of the reduction parameters (S,r,s): \n";
    for(unsigned int iterator=0; iterator<indices.size();iterator++){
      logger << indices[iterator] << " ";
    }
    logger << "\n";
    sqlite3_reset(stmt);
    return 0;
  }
  
  
  if(something_int(sqlite3_column_text(stmt, 7))!=0){
    logger << "This integral is not inside the range of the reduction parameters (S,r,s) and \n";
    for(unsigned int iterator=0; iterator<indices.size();iterator++){
      logger << indices[iterator] << " ";
    }
    logger << "\n";
    sqlite3_reset(stmt);
    return 0;
  }
  
  unsigned long id;
  std::string unsignedLong = (char*)sqlite3_column_text(stmt, 0);
  std::istringstream ss(unsignedLong);
  ss >> id;

  sqlite3_reset(stmt);
  return id;
}